import enum
import EER_JSON
import ARM_JSON

# Constants
OID = 'OID'
FOREIGN_KEY_NULL = True
# Set to False to force mandatory participation of an Entity type in a binary relationship.

# Constants for candidate maps.
PRIMARY_KEY = 'PrimaryKey'
RELATION_TYPE = 'RelationType'
PARTICIPANTS_KEY = 'ParticipantsKey'
CANDIDATE_TYPE = 'CandidateType'
EER_TYPE = 'EERType'

# Helper methods

class CandidateType(enum.Enum):
    STRONG_ENTITY = enum.auto()
    WEAK_ENTITY = enum.auto()
    PARTIAL_ENTITY = enum.auto()
    REGULAR_RELATION = enum.auto()
    SUBTYPE = enum.auto()


def get_primary_key(arm_relation_type):
    constraints = ARM_JSON.constraints(arm_relation_type)
    primary_key_attribute = ARM_JSON.primary_key_attribute(ARM_JSON.primary_key_constraint(constraints))
    path_fd_constraints = ARM_JSON.pathfd_constraints(constraints)
    identifying_pathfd = None
    for path_fd_constraint in path_fd_constraints:
        if ARM_JSON.is_identifying(path_fd_constraint, primary_key_attribute):
            identifying_pathfd = path_fd_constraint
            break
    if identifying_pathfd:
        path_fd_constraints.remove(identifying_pathfd)
    return identifying_pathfd, path_fd_constraints

def classify_attributes(pathfd_constraint, arm_relation_type):
        foreign_key_attributes = ARM_JSON.foreign_key_attributes(arm_relation_type)
        local = list()
        foreign = list()
        paths = ARM_JSON.constraint_paths(pathfd_constraint)
        for path in paths:
            attribute_name = ARM_JSON.attribute_name(path)
            if attribute_name in foreign_key_attributes:
                assert(ARM_JSON.get_data_type(arm_relation_type, attribute_name) == OID)
                foreign.append(attribute_name)
            else:
                local.append(attribute_name)
        return local, foreign

	
def local_key_attributes(pathfd_constraint, arm_relation_type):
    return classify_attributes(pathfd_constraint, arm_relation_type)[0]

def foreign_key_attributes(pathfd_constraint, arm_relation_type):
    return classify_attributes(pathfd_constraint, arm_relation_type)[1]


def make_binary_relationship(arm_relation_type, name, foreign_key_constraints):
    relation_name = ARM_JSON.relation_type_name(arm_relation_type)
    for foreign_key_constraint in foreign_key_constraints:
        constraint_name, attribute_name, foreign_relation_name = ARM_JSON.foreign_key_components(foreign_key_constraint)
        if name==attribute_name:
            # relationship name composed of attribute name and target relation
            relationship_type =  EER_JSON.create_relationship_type(name+'_'+foreign_relation_name)
            EER_JSON.add_participant(relationship_type, relation_name)

            participation = '0' if FOREIGN_KEY_NULL else '1'
            EER_JSON.add_participant(relationship_type, foreign_relation_name, min_cardinality = participation, max_cardinality = '1')
            return relationship_type
    raise Exception("ARM_To_EER:make_binary_relationship: Failed to find foreign key constraint for given attribute name.")


def classify_relations(arm_model, log):
    entity_type_candidates = list()  # list of dictionaries storing data on ARM relations identified as entity types. DUPLICATES?
    relationship_type_candidates = list()  # List of dictionaries storing  data on ARM relations identified as relationship types. DUPLICATES?
    for arm_relation_type in ARM_JSON.get_relation_types(arm_model):
        # I.A. Select primary key
        relation_type_name = ARM_JSON.relation_type_name(arm_relation_type)

        primary_key, rejects = get_primary_key(arm_relation_type)
        if primary_key is None:
            # Is this a subtype?
            if ARM_JSON.is_subtype(arm_relation_type):
                log.append("Relation type \'{:s}\' mapped to EER entity sub type \'{:s}\'.".format(relation_type_name, relation_type_name))
                candidate = { # This ARM relation type maps to an entity type subtype with no identity of its own.
                    CANDIDATE_TYPE: CandidateType.SUBTYPE,
                    RELATION_TYPE: arm_relation_type,
                    PRIMARY_KEY: None
                }
                entity_type_candidates.append(candidate)
            else:
                log.append('Error: Relation type \'{:s}\' has no identifying pathfd, tentatively mapped to EER entity type.'.format(relation_type_name))
                # Add it anyway
                candidate = { # This ARM relation type has no primary key and is not a subtype.
                    CANDIDATE_TYPE: CandidateType.PARTIAL_ENTITY,
                    RELATION_TYPE: arm_relation_type,
                    PRIMARY_KEY: None
                }
                entity_type_candidates.append(candidate)
        else:
            local_attributes, foreign_attributes = classify_attributes(primary_key, arm_relation_type)
            if not foreign_attributes:
                # I.B. 'strong Entity'
                log.append("Relation type \'{:s}\' mapped to strong entity type \'{:s}\'.".format(relation_type_name, relation_type_name))
                candidate = { # This ARM relation type maps to a strong entity type
                    CANDIDATE_TYPE: CandidateType.STRONG_ENTITY,
                    RELATION_TYPE: arm_relation_type,
                    PRIMARY_KEY: primary_key
                }
                entity_type_candidates.append(candidate)
            elif not local_attributes:
                # I.C. 'regular relation'.
                log.append("Relation type \'{:s}\' mapped to regular relationship type \'{:s}\'.".format(relation_type_name, relation_type_name))
                candidate = { # This ARM relation type maps to a regular relationship type
                    CANDIDATE_TYPE: CandidateType.REGULAR_RELATION,
                    RELATION_TYPE: arm_relation_type,
                    PRIMARY_KEY: None,
                    PARTICIPANTS_KEY: primary_key
                }
                relationship_type_candidates.append(candidate)
            else:
                # I.D. 'weak relation'
                # A relation with primary key containing some but not all foreign key attributes.
                log.append("Relation type \'{:s}\' mapped to weak entity type \'{:s}\'.".format(relation_type_name, relation_type_name))
                candidate = { # This ARM relation type maps to a weak entity type.
                    CANDIDATE_TYPE: CandidateType.WEAK_ENTITY,
                    RELATION_TYPE: arm_relation_type,
                    PRIMARY_KEY: primary_key
                }
                entity_type_candidates.append(candidate)
        # Update log with unused pathfds
        for path_fd_constraint in rejects:
            log.append("The constraint \'{:s}\' for relation type \'{:s}\' is not identifying and not translatable.".format(ARM_JSON.pathfd_string(path_fd_constraint), relation_type_name))

    return entity_type_candidates, relationship_type_candidates


def initial_eer_model(entity_type_candidates, relationship_type_candidates, log):
    eer_model = EER_JSON.create_eer_model()

    for candidate in entity_type_candidates:
        arm_relation_type = candidate[RELATION_TYPE]
        eer_entity_type = EER_JSON.add_entity_type(eer_model, ARM_JSON.relation_type_name(arm_relation_type))
        candidate[EER_TYPE] = eer_entity_type

    for candidate in relationship_type_candidates:
        arm_relation_type = candidate[RELATION_TYPE]
        arm_relation_name = ARM_JSON.relation_type_name(arm_relation_type)
        foreign_keys = foreign_key_attributes(candidate[PARTICIPANTS_KEY], arm_relation_type)

        eer_relationship_type = EER_JSON.add_relationship_type(eer_model, arm_relation_name)
        candidate[EER_TYPE] = eer_relationship_type

        # Complete relation (step I.B./step I.C.) by adding participation of relevant entity types
        log.append('Processed foreign key constraints of ARM relation \'{:s}\'.'.format(ARM_JSON.relation_type_name(arm_relation_type)))
        foreign_key_constraints = ARM_JSON.foreign_key_constraints(ARM_JSON.constraints(arm_relation_type))
        for foreign_key_constraint in foreign_key_constraints:
            constraint_name, attribute_name, relation_name = ARM_JSON.foreign_key_components(foreign_key_constraint)
            assert(attribute_name in foreign_keys)
            foreign_keys.remove(attribute_name)
            log.append('- added entity type \'{:s}\' to the EER relationship \'{:s}\'.'.format(relation_name, EER_JSON.name(eer_relationship_type)))
            EER_JSON.add_participant(eer_relationship_type, relation_name, role=constraint_name, participation=EER_JSON.Participation.STRONG)

    return eer_model


def configure_identities(candidates, log):
    for candidate in candidates:
        primary_key = candidate[PRIMARY_KEY]
        primary_key_attribute = ARM_JSON.primary_key_attribute(ARM_JSON.primary_key_constraint(\
            ARM_JSON.constraints(candidate[RELATION_TYPE])))
        if primary_key is not None:  # Not a subtype (or underspecified relation type.), or regular relation
            eer_type  = candidate[EER_TYPE]
            # Add attributes that form part of identity
            log.append('Derived identity for EER type \'{:s}\' based on constraints \'primary key ({:s})\' and \'{:s}\' of the corresponding ARM relation.'\
                       .format(EER_JSON.name(eer_type), primary_key_attribute, ARM_JSON.constraint_to_string(primary_key)))
            local_attributes, foreign_key_attributes = classify_attributes(primary_key, candidate[RELATION_TYPE])
            local_attributes.sort()
            for attribute_name in local_attributes:
                EER_JSON.add_attribute(eer_type, attribute_name)
            # Add description of identity
            foreign_key_attributes.sort()
            if foreign_key_attributes:
                # Partial/Weak identity
                EER_JSON.set_identity(eer_type, local_attributes, EER_JSON.IdentityType.PARTIAL)
            else:
                # Complete/Strong identity
                EER_JSON.set_identity(eer_type, local_attributes, EER_JSON.IdentityType.COMPLETE)


def find_candidate(candidates, name):
    for candidate in candidates:
        if ARM_JSON.relation_type_name(candidate[RELATION_TYPE]) == name:
            return candidate
    return None

def process_non_identifying_attributes(eer_model, candidates, log):
    # Very clunky and needs a rewrite
    for candidate in candidates:
        arm_relation_type = candidate[RELATION_TYPE]
        fk_attributes = ARM_JSON.foreign_key_attributes(arm_relation_type)
        non_fk_attributes = ARM_JSON.non_fk_attributes(arm_relation_type)
		
        eer_type  = candidate[EER_TYPE]
        eer_type_name = EER_JSON.name(eer_type)
        identity = EER_JSON.identity(eer_type)

        log.append('Processed non-identifying attributes of ARM relation \'{:s}\'.'.format(eer_type_name, eer_type_name))

        non_ident_attribs = non_fk_attributes.keys()
        if identity is not None: # A Strong or weak entity
            non_ident_attribs = non_ident_attribs-EER_JSON.identifying_attributes(identity)
            # Pull the ARM relation identifying attribute name from the primary key constraint.

        primary_key_attribute = ARM_JSON.primary_key_attribute(ARM_JSON.primary_key_constraint(\
            ARM_JSON.constraints(candidate[RELATION_TYPE])))
        if primary_key_attribute is not None:
            non_ident_attribs = non_ident_attribs-{primary_key_attribute}
            log.append('- dropped \'{} OID\'.'.format(primary_key_attribute))

        # Process remaining non fk attributes in arm relation
        non_ident_attribs = list(non_ident_attribs)
        non_ident_attribs.sort()
        for name in non_ident_attribs:
            EER_JSON.add_attribute(eer_type, name)
            log.append('- mapped attribute \'{:s}\', discarding datatype.'.format(name))

        # Process remaining foreign attributes, adding EER type to existing relationships or creating new ones.
        binary_relationships = []
        foreign_key_constraints = ARM_JSON.foreign_key_constraints(ARM_JSON.constraints(arm_relation_type))
        if candidate[CANDIDATE_TYPE] == CandidateType.REGULAR_RELATION:
            # It's a relationship, we need to remove from consideration the foreign key attributes used to build it.
            fk_attribute_names = fk_attributes.keys()-set(foreign_key_attributes(candidate[PARTICIPANTS_KEY], arm_relation_type))
            # Result ought to be empty...
        else:
            fk_attribute_names = fk_attributes.keys()

        if fk_attribute_names:
            log.append('Processed foreign key constraints of ARM relation \'{:s}\'.'.format(ARM_JSON.relation_type_name(arm_relation_type)))
            for constraint in foreign_key_constraints:
                constraint_name, foreign_key_name, foreign_relation_name = ARM_JSON.foreign_key_components(constraint)

                if foreign_key_name in fk_attribute_names:
                    # Located the constraint for a remaining FK attribute
                    if candidate[CANDIDATE_TYPE] == CandidateType.REGULAR_RELATION:
                        # Exclude ARM relations that are mapped to relationship types.
                        log.append('Error: Relation type \'{:s}\' mapped to an EER relationship type but is also a participant in a relationship with \'{:s}\', dropped participation.'.format(eer_type_name, ARM_JSON.foreign_key_relation(constraint)))

                    if EER_JSON.has_relationship_type(eer_model, foreign_relation_name):
                        # Foreign ARM relation has been mapped to an EER relationship
                        # Can add this EER type as a participant in an existing relationship
                        log.append('- added entity type \'{:s}\' to relationship \'{:s}\'.'.format(eer_type_name, foreign_relation_name))
                        target_relationship = EER_JSON.get_relationship_type(eer_model, foreign_relation_name)
                    else:
                        # Foreign ARM relation has been mapped to an entity type.
                        # Construct a new (binary) relationship, naming it after the fk attribute and the foreign relation,
                        # Add eer_type and foreign entity type as participants
                        assert(EER_JSON.has_entity_type(eer_model, foreign_relation_name))
                        relationship_name = constraint_name.capitalize() # Should really check for uniqueness before using as relationship name
                        log.append('- added \'{:s}, {:s}\' to new relationship \'{:s}\'.'.format(eer_type_name, foreign_relation_name, relationship_name))
                        target_relationship = EER_JSON.add_relationship_type(eer_model, relationship_name)

                    foreign_candidate = find_candidate(candidates, foreign_relation_name)
                    assert(foreign_candidate!=None and foreign_candidate[CANDIDATE_TYPE] != CandidateType.WEAK_ENTITY)
                    EER_JSON.add_participant(target_relationship, foreign_relation_name, participation=EER_JSON.Participation.STRONG) # Don't wish to set cardinality as don't know intention.
                    # NEED VALIDATION HERE !!!!

                    # Determine whether this EER type depends on the  relationship
                    if candidate[CANDIDATE_TYPE] == CandidateType.WEAK_ENTITY and\
                            foreign_key_name in foreign_key_attributes(candidate[PRIMARY_KEY], candidate[RELATION_TYPE]):
                        participation_type = EER_JSON.Participation.WEAK
                        participation = '1'
                    else:
                        participation_type = EER_JSON.Participation.STRONG
                        participation = '0' if FOREIGN_KEY_NULL else '1'

                    EER_JSON.add_participant(target_relationship, eer_type_name,
                                             min_cardinality=participation, max_cardinality='1', participation=participation_type)



def process_specialisations(eer_model, candidates, log):
    specialisation_candidates = {}
    # Build a sub-type map
    super_to_sub = {} # Need this to weed out subtypes not in coverage constraint.
    sub_to_super = {}
    super_to_coverage = {}
    disjoint_to_super = {}
    disjointness_constraints = {}

    # Populate data structures.
    for candidate in candidates:
        arm_relation_type = candidate[RELATION_TYPE]
        relation_type_name = ARM_JSON.relation_type_name(arm_relation_type)
        isa_constraint_list = ARM_JSON.specialisation_constraints(ARM_JSON.constraints(arm_relation_type))
        disjoint_with_constraint_list = ARM_JSON.disjointness_constraints(ARM_JSON.constraints(arm_relation_type))
        coverage_constraint_list = ARM_JSON.coverage_constraints(ARM_JSON.constraints(arm_relation_type))
        # Check the candidate hasn't been mapped to a relationship
        if (isa_constraint_list or coverage_constraint_list) \
            and candidate[CANDIDATE_TYPE] == CandidateType.REGULAR_RELATION:
            log.append('Error: ARM relation type \'{:s}\' maps to an EER relationship type but it contains the following specialisation-related constraints.')
            for constraint in isa_constraint_list.union(coverage_constraint_list):
                log.append('- \'{:s}\''.format(ARM_JSON.constraint_to_string(constraint)))
            log.append('The constraints were dropped.')
            break
        else:
            if isa_constraint_list:
                # Build map of supertype to subtypes and subtypes to super types.
                specialisation_candidates[relation_type_name] = candidate
                sub_to_super[relation_type_name] = set()
                for constraint in isa_constraint_list:
                    super_type = ARM_JSON.constraint_supertype(constraint)
                    sub_type = relation_type_name
                    if not super_type in super_to_sub.keys():
                        # Use a set to eradicate duplicates
                        super_to_sub[super_type] = set()
                    super_to_sub[super_type].add(sub_type)
                    sub_to_super[relation_type_name].add(super_type)

            if coverage_constraint_list:
                # Build a map of relation name to coverage constraints.
                for constraint in coverage_constraint_list:
                    # Assume that coverage constraints are disjoint.
                    # Assume that the referenced relations are subtypes of the relation containing the constraint.
                    if relation_type_name not in super_to_coverage.keys():
                        super_to_coverage[relation_type_name] = list()
                    log.append('Defined a specialisation for entity type \'{:s}\' based on coverage constraint \'{:s}\' in ARM relation \'{:s}\'.'.format(relation_type_name, ARM_JSON.constraint_to_string(constraint), relation_type_name))

                    super_to_coverage[relation_type_name].append(frozenset(ARM_JSON.constraint_subtypes(constraint)))

            if disjoint_with_constraint_list:
                # Build a map of constraint relation sets to constraints.
                for constraint in disjoint_with_constraint_list:
                    # Set of relations referenced in constraint
                    relation_type_set = frozenset(ARM_JSON.constraint_relations(constraint)+[relation_type_name])
                    if relation_type_set not in disjointness_constraints:
                        disjointness_constraints[relation_type_set] = [{'relation': relation_type_name, 'constraint': constraint}]
                    else:
                        disjointness_constraints[relation_type_set].append({'relation': relation_type_name, 'constraint': constraint})
    # From disjointness_constraints map, identify sets that qualify as associated with a specialisation
    # Check symmetry. e.g. Given a disjoint with b,c, should also have b disjoint with a,c, and c disjoint with a, b.
    # Check that the relations involved in the constraint have a common supertype.
    specialisation_disjoints = {}
    for relation_type_set in disjointness_constraints.keys():
        qualify = True
        if len(relation_type_set) != len(disjointness_constraints[relation_type_set]):
            # Symmetry requirement not satisfied.
            qualify = False
        else:
            # Check there's a common supertype.
            iterator=relation_type_set.__iter__()
            relation_type_name = next(iterator)
            if relation_type_name not in sub_to_super:
                # it's not a sub type
                qualify = False
            else:
                supertype_set = sub_to_super[relation_type_name]
                while True:
                    try:
                        relation_type_name = next(iterator)
                        if relation_type_name in sub_to_super:
                            supertype_set = supertype_set.intersection(sub_to_super[relation_type_name])
                        else:
                            supertype_set = set()
                            qualify = False
                            break
                    except StopIteration:
                        break
                if supertype_set:
                    # There's at least one supertype in common, so this constraint set can map to a specialisation.
                    for supertype in supertype_set:
                        if supertype not in specialisation_disjoints:
                            specialisation_disjoints[supertype] = list()
                        if relation_type_set not in specialisation_disjoints[supertype]:
                            specialisation_disjoints[supertype].append(relation_type_set)
        if not qualify:
            for constraint_object in disjointness_constraints[relation_type_set]:
                log.append('Relation type \'{:s}\' has constraint \'{:s}\' but constraint discarded as could not map to a specialisation.'.format(constraint_object['relation'], ARM_JSON.constraint_to_string(constraint_object['constraint'])))

    # Start by building specialisations based on coverage constraints.
    specialisations = {}
    # Build a sorted list and thus a deterministic output to facilitate testing
    supertype_list = list(super_to_coverage.keys())
    supertype_list.sort()

    for super in supertype_list:
        specialisations[super] = list()
        for covered_subtype_set in super_to_coverage[super]:
            assert(covered_subtype_set.issubset(super_to_sub[super]))
            specialisation = EER_JSON.add_specialisation(eer_model, super, covered_subtype_set)
            super_to_sub[super] = super_to_sub[super]-covered_subtype_set
            if not super_to_sub[super]:
                super_to_sub.pop(super)

            EER_JSON.add_constraint(specialisation, EER_JSON.Coverage.COMPLETE)

            # See if there's a corresponding disjointness requlrement
            if super in specialisation_disjoints and covered_subtype_set in specialisation_disjoints[super]:
                log.append('Identified disjointness requirement for the specialisation \'({:s})\' of \'{:s}\'.'.format(ARM_JSON.join(sorted(covered_subtype_set), ', '),super))
                specialisation_disjoints[super].remove(covered_subtype_set)
                if not specialisation_disjoints[super]:
                    specialisation_disjoints.pop(super)
                EER_JSON.add_constraint(specialisation, EER_JSON.Disjointness.DISJOINT)

            # Assert that there's no overlap in the specialisations for this supertype....
            specialisations[super].append(specialisation)

    # Deal with the rest of the subtype disjointness constraints.
    # Build a sorted list and thus a deterministic output to facilitate testing
    supertypes = list(specialisation_disjoints.keys())
    supertypes.sort()
    for super in supertypes:
        if not super in specialisations:
            specialisations[super] = list()
        for disjoint_set in specialisation_disjoints[super]:
            if (not super_to_sub[super] or not disjoint_set.issubset(super_to_sub[super])):
                log.append("Relation type \'{:s}\' has disjoint sub type constraints for \'{:s}\' but constraints discarded as could not map to a specialisation.".format(super, '('+ARM_JSON.join(sorted(disjoint_set), ', ')+')'))
            else:
                log.append('Defined a specialisation \'({:s})\' of \'{:s}\' based on disjointness constraints in the corresponding ARM relations.'.format(ARM_JSON.join(sorted(disjoint_set), ', '), super))
                specialisation = EER_JSON.add_specialisation(eer_model, super, disjoint_set)
                super_to_sub[super] = super_to_sub[super]-((disjoint_set))
                if not super_to_sub[super]:
                    super_to_sub.pop(super)
                EER_JSON.add_constraint(specialisation, EER_JSON.Disjointness.DISJOINT)
                specialisations[super].append(specialisation)

    # Deal with remaining subtypes that are not grouped under a completeness/disjointness constraint.
    supertypes = list(super_to_sub.keys())
    supertypes.sort()
    for super in supertypes:
        if not super in specialisations:
            specialisations[super] = list()
        log.append('Defined the specialisation \'({:s})\' of type \'{:s}\'.'.format(ARM_JSON.join(sorted(super_to_sub[super]), ', '), super))
        specialisation = EER_JSON.add_specialisation(eer_model, super, super_to_sub[super])
        specialisations[super].append(specialisation)


def transform(arm_model):
    log = ["TRANSFORMATION REPORT"]

    # I. Build the initial ERD
    # Steps I.A to I.D classify the ARM relation types
    entity_type_candidates, relationship_type_candidates = classify_relations(arm_model, log)
    all_candidates = entity_type_candidates+relationship_type_candidates

    # Build the initial model.
    eer_model = initial_eer_model(entity_type_candidates, relationship_type_candidates, log)

    configure_identities(all_candidates, log)
#    for candidate in all_candidates:
#        # Step II Configure Entity type identities
#        configure_identity(candidate, log)

        # Step III Process remaining attributes (those not in primary key)
    process_non_identifying_attributes(eer_model, all_candidates, log)

    # Handle specialisations (steps IV-VI)
    process_specialisations(eer_model, all_candidates, log)

    EER_JSON.add_notes(eer_model, log)
    return eer_model
