import enum
"""
Wrapper module for JSON EER model
"""

# Constants
UNKNOWN = 'unknown'

# Operations on EER MODEL
DOMAINS = 'domains'
ENTITY_TYPES = 'entityTypes'
RELATIONSHIP_TYPES = 'relationshipTypes'
SPECIALISATIONS = 'specialisations'
NOTES = 'notes'

def has_domains(eer_model):
    return DOMAINS in eer_model

def domains(eer_model):
    assert(DOMAINS in eer_model)
    return eer_model[DOMAINS]

def add_domain(eer_model, domain):
    if not DOMAINS in eer_model:
        eer_model[DOMAINS] = []
    eer_model[DOMAINS].append(domain)

def entity_types(eer_model):
    assert(ENTITY_TYPES in eer_model)
    return eer_model[ENTITY_TYPES]


def has_entity_type(eer_model, entity_type_name):
    return get_entity_type(eer_model, entity_type_name) is not None


def get_entity_type(eer_model, entity_type_name):
    for entity_type in entity_types(eer_model):
        if name(entity_type)==entity_type_name:
            return entity_type
    return None


def has_relationship_types(eer_model):
    return RELATIONSHIP_TYPES in eer_model


def relationship_types(eer_model):
    assert(RELATIONSHIP_TYPES in eer_model)
    return eer_model[RELATIONSHIP_TYPES]


def has_relationship_type(eer_model, relationship_type_name):
    return has_relationship_types(eer_model) and get_relationship_type(eer_model, relationship_type_name) is not None


def get_relationship_type(eer_model, relationship_type_name):
    for relationship_type in relationship_types(eer_model):
        if name(relationship_type)==relationship_type_name:
            return relationship_type
    return None


def has_specialisations(eer_model):
    return SPECIALISATIONS in eer_model

def get_specialisations(eer_model):
    assert(SPECIALISATIONS in eer_model)
    return eer_model[SPECIALISATIONS]

def has_notes(eer_model):
    return NOTES in eer_model

def get_notes(eer_model):
    return eer_model[NOTES]


def create_eer_model():
    return {ENTITY_TYPES: list()}


def add_entity_type(eer_model, entity_type_name):
    assert(ENTITY_TYPES in eer_model)
    entity_type = create_entity_type(entity_type_name)
    eer_model[ENTITY_TYPES].append(entity_type)
    return entity_type


def add_relationship_type(eer_model, relationship_type_name):
    relationship_type = create_relationship_type(relationship_type_name)
    if RELATIONSHIP_TYPES not in eer_model:
        eer_model[RELATIONSHIP_TYPES] = list()
    eer_model[RELATIONSHIP_TYPES].append(relationship_type)
    return relationship_type


def add_specialisation(eer_model, supertype_name, subtype_names):
    specialisation = create_specialisation(supertype_name, subtype_names)
    if SPECIALISATIONS not in eer_model:
        eer_model[SPECIALISATIONS] = list()
    eer_model[SPECIALISATIONS].append(specialisation)
    return specialisation


def add_notes(eer_model, notes):
    eer_model[NOTES] = notes


# Operations common to more than one model component
ATTRIBUTES = 'attributes'
NAME = 'name'


def name(json_map): # for entity type, attribute, relationship type, subtype
    assert(NAME in json_map)
    return json_map[NAME]


def has_attributes(json_map): # for entity type, relationship type
    return ATTRIBUTES in json_map


def attributes(json_map): # for entity type, relationship type
    assert(ATTRIBUTES in json_map)
    return json_map[ATTRIBUTES]


def attribute_names(json_map):
    assert(ATTRIBUTES in json_map)
    return { attribute[NAME] for attribute in attributes(json_map)}



def add_attribute(json_map, name, dataType=None):
    if ATTRIBUTES not in json_map:
        json_map[ATTRIBUTES] = []
    if dataType is not None:
        json_map[ATTRIBUTES].append({NAME : name, DATA_TYPE : dataType})
    else:
        json_map[ATTRIBUTES].append({NAME : name})


# Operations on Entity Type
IDENTITY = 'identity'

def has_identity(json_map): # for entity type
    return IDENTITY in json_map


def identity(json_map): # for entity typ
    if IDENTITY in json_map:
        return json_map[IDENTITY]
    else:
        return None


def set_identity(json_map, attribute_names, identity_type): # for entity type
    json_map[IDENTITY] = create_identity(attribute_names, identity_type)


def create_entity_type(entity_type_name):
    return { NAME: entity_type_name, ATTRIBUTES: list()}


def weak(entity_type):
    assert(IDENTITY in entity_type)
    return identity_type(entity_type[IDENTITY])==IdentityType.PARTIAL


# Operations on attributes
DATA_TYPE = 'dataType'
UNKNoWN = 'unknown'


def has_data_type(attribute):
    return DATA_TYPE in attribute


def get_data_type(attribute):
    assert(DATA_TYPE in attribute)
    return attribute[DATA_TYPE]


def is_compound(data_type):
    return type(data_type) != str


# Operations on Identity
IDENTITY_TYPE = 'type'
ATTRIBUTE_NAMES = 'attributes'


class IdentityType(enum.Enum):
    PARTIAL = 'partial'
    COMPLETE = 'complete'

    @classmethod
    def parse(cls, string):
        for value in IdentityType:
            if value.value==string:
                return value
        return None

def identity_type(identity):
    assert(IDENTITY_TYPE in identity)
    assert(IdentityType.parse(identity[IDENTITY_TYPE]) is not None)
    return IdentityType.parse(identity[IDENTITY_TYPE])


def identifying_attributes(identity):
    assert(ATTRIBUTE_NAMES in identity)
    return set(identity[ATTRIBUTE_NAMES])


def create_identity(attribute_names, identity_type):
    return { ATTRIBUTE_NAMES: attribute_names, IDENTITY_TYPE: identity_type.name.lower()}


# Operations on Specialisation
SUPERTYPE = 'supertype'
SUBTYPES = 'subtypes'
DISJOINTNESS = 'disjointness'
COVERAGE = 'coverage'
PREDICATE = 'predicate'

class Coverage(enum.Enum):
    PARTIAL = 'partial'
    COMPLETE = 'complete'

    @classmethod
    def parse(cls, string):
        for value in Coverage:
            if value.value==string:
                return value
        return None


class Disjointness(enum.Enum):
    OVERLAPPING = 'overlapping'
    DISJOINT = 'disjoint'

    @classmethod
    def parse(cls, string):
        for value in Disjointness:
            if value.value==string:
                return value
        return None

def get_supertype(specialisation):
    assert(SUPERTYPE in specialisation)
    return specialisation[SUPERTYPE]


def get_subtypes(specialisation):
    assert (SUBTYPES in specialisation)
    return specialisation[SUBTYPES]


def has_constraints(specialisation):
    return COVERAGE or DISJOINTNESS in specialisation


def has_coverage_constraint(specialisation):
    return COVERAGE in specialisation


def get_coverage_constraint(specialisation):
    assert(COVERAGE in specialisation)
    value = specialisation[COVERAGE]
    assert(Coverage.parse(value) is not None)
    return Coverage.parse(value)


def has_disjointness_constraint(specialisation):
    return DISJOINTNESS in specialisation


def get_disjointness_constraint(specialisation):
    assert(DISJOINTNESS in specialisation)
    value = specialisation[DISJOINTNESS]
    assert(Disjointness.parse(value) is not None)
    return Disjointness.parse(value)


def create_specialisation(super_type_name, subtype_names):
    subtype_names = list(subtype_names)
    subtype_names.sort()
    specialisation = { SUPERTYPE: super_type_name, SUBTYPES: subtype_names}
    return specialisation

def add_constraint(specialisation, constraint_value):
    assert(SUPERTYPE in specialisation and SUBTYPES in specialisation)
    if constraint_value in Coverage:
        specialisation[COVERAGE] = constraint_value.value
    elif constraint_value in Disjointness:
        specialisation[DISJOINTNESS] = constraint_value.value


# Operations on Relationships
PARTICIPANTS = 'participants'


def get_participants(relationship_type):
    assert(PARTICIPANTS in relationship_type)
    return relationship_type[PARTICIPANTS]


def create_relationship_type(relationship_type_name):
    return { NAME: relationship_type_name, PARTICIPANTS: list() }


def add_participant(relationship_type, name, role=None, min_cardinality=None, max_cardinality=None, participation=None):
    assert(PARTICIPANTS in relationship_type)
    participant = create_participant(name)
    if role is not None:
        set_role(participant, role)
    if min_cardinality is not None:
        set_min_cardinality(participant, min_cardinality)
    if max_cardinality is not None:
        set_max_cardinality(participant, max_cardinality)
    if participation is not None:
        set_participation(participant, participation)
    relationship_type[PARTICIPANTS].append(participant)

def is_identifying(relationship_type):
    assert(PARTICIPANTS in relationship_type)
    for participant in relationship_type[PARTICIPANTS]:
        if has_participation(participant) and get_participation(participant)==Participation.WEAK.value:
            return True
    return False

def is_binary(relationhip_type):
    assert(PARTICIPANTS in relationhip_type)
    return len(get_participants(relationhip_type)) == 2

def is_many_to_many(relationship_type):
    assert(is_binary(relationship_type))
    participants = get_participants(relationship_type)
    return cardinality_is_many(participants[0]) and cardinality_is_many(participants[1])

def is_one_to_many(relationship_type):
    assert(is_binary(relationship_type))
    participants = get_participants(relationship_type)
    result = cardinality_is_many(participants[0]) and cardinality_is_one(participants[1])
    return result or cardinality_is_one(participants[0]) and cardinality_is_many(participants[1])

# Operations on participants
PARTICIPANT_ROLE = 'role'
CARDINALITY_MIN = 'min' # Assuming look-there
CARDINALITY_MAX = 'max' # cardinality constraints.
PARTICIPATION = 'participation'


class Participation(enum.Enum):
    WEAK = 'weak'
    STRONG = 'strong'

    @classmethod
    def parse(cls, string):
        for value in Participation:
            if value.value==string:
                return value
        return None

def has_participation(participant):
    return PARTICIPATION in participant


def get_participation(participant):
    assert(Participation.parse(participant[PARTICIPATION]) is not None)
    return  Participation.parse(participant[PARTICIPATION])


def has_role(participant):
    return PARTICIPANT_ROLE in participant


def get_role(participant):
    return participant[PARTICIPANT_ROLE]


def has_min_cardinality(participant):
    return CARDINALITY_MIN in participant


def has_max_cardinality(participant):
    return CARDINALITY_MAX in participant


def get_min_cardinality(participant):
    assert(CARDINALITY_MIN in participant)
    return participant[CARDINALITY_MIN].strip()


def get_max_cardinality(participant):
    assert(CARDINALITY_MIN in participant)
    return participant[CARDINALITY_MAX].strip()


def cardinality_is_many(participant):
    assert(has_max_cardinality(participant))
    max = get_max_cardinality(participant)
    return max not in {'0', '1'}


def cardinality_is_one(participant):
    assert(has_max_cardinality(participant))
    max = get_max_cardinality(participant)
    return max == '1'


def optional_participant(participant):
    return has_min_cardinality(participant) and get_min_cardinality(participant)=='0'


def create_participant(name):
    return { NAME: name}


def set_participation(participant, participation):
    participant[PARTICIPATION] = participation.value


def set_role(participant, role):
    participant[PARTICIPANT_ROLE] = role


def set_min_cardinality(participant, value):
    participant[CARDINALITY_MIN]=value


def set_max_cardinality(participant, value):
    participant[CARDINALITY_MAX] = value

