import enum
import EER_JSON
import ARM_JSON

# Constants
FOREIGN_KEY_NULL = True
# Set to False to force binary relations with optional participation to be modeled as relations.
# Set to True to permit binary relations with optional participation to be modeled as foreign keys.

# Helper methods
def sortedList(set):
    result = list(set)
    result.sort()
    return result

# Composite attributes....
# Make a relation with composite attributes. Give it a self, Put FK in relevant entities/relations.
# Keep a dictionary of composites and match other occurrences.


def transfer_attributes(eer_type, arm_relation_type, log, prefix=None):
    assert(EER_JSON.has_attributes(eer_type))
    for attribute in EER_JSON.attributes(eer_type):
        attribute_name = EER_JSON.name(attribute)
        if prefix:
            attribute_name = prefix+'_'+attribute_name
        # Avoid name clashes
        if attribute_name == ARM_JSON.SELF:
            attribute_name = attribute_name+'*'
            log.append('- attribute \'{:s}.self\' renamed to \'{:s}\' to avoid clash in ARM model.'.format(EER_JSON.name(eer_type), attribute_name))

        if EER_JSON.has_data_type(attribute):
            attribute_data_type = EER_JSON.get_data_type(attribute)
            if EER_JSON.is_compound(attribute_data_type):
                log.append('- attribute \'{:s}\' is composite. Translation not implemented. Marked datatype as \'anyType\'.'.format(attribute_name))
                ARM_JSON.add_attribute(arm_relation_type, attribute_name)
                pass
            else:
                log.append('- attribute \'{:s}\' has datatype \'{:s}\', transferred it to ARM model.'.format(attribute_name, attribute_data_type))
                ARM_JSON.add_attribute(arm_relation_type, attribute_name, attribute_data_type)
        else:
            log.append('- attribute \'{:s}.{:s}\' has no datatype. Marked as \'anyType\'.'.format(EER_JSON.name(eer_type), EER_JSON.name(attribute)))
            ARM_JSON.add_attribute(arm_relation_type, attribute_name)
    return


def transfer_identity(eer_type, arm_relation_type, log):
    assert( ARM_JSON.SELF in ARM_JSON.attribute_names(arm_relation_type) )
    attribute_names = EER_JSON.identifying_attributes(EER_JSON.identity(eer_type))
    log.append('- added \'pathfd({:s})->self\' constraint created from identity.'.format(ARM_JSON.join(sorted(attribute_names), ', ')))
    ARM_JSON.add_pathfd_constraint(arm_relation_type, ARM_JSON.create_simple_paths(attribute_names), ARM_JSON.create_simple_path(ARM_JSON.SELF))
    # Note, in the case of a weak identity, the identifying pathfd constraint will be augmented with foreign keys as
    # identifying relationships are processed.
    return


def transfer_entity_types(eer_model, arm_model, log):
    for entity_type in EER_JSON.entity_types(eer_model):
        entity_type_name = EER_JSON.name(entity_type)
        log.append("EER entity type \'"+entity_type_name+"\': transformed to ARM relation type \'"+entity_type_name+"\'.")
        arm_relation_type = ARM_JSON.add_relation_type(arm_model, entity_type_name)
        transfer_attributes(entity_type, arm_relation_type, log)
        # include 'self'
        log.append('- added \'self\' attribute.')
        ARM_JSON.add_attribute(arm_relation_type, ARM_JSON.SELF, ARM_JSON.OID)
        if EER_JSON.has_identity(entity_type): # Subtypes don't.
            transfer_identity(entity_type, arm_relation_type, log)
        log.append('- added \'primary key (self)\' constraint.')
        ARM_JSON.add_primary_key_constraint(arm_relation_type, ARM_JSON.SELF)

    return

def transfer_specialisations(eer_model, arm_model, log):
    log.append('Specialisations mapped to \'isa\', \'covered by\' and \'disjoint with\' constraints.')
    for specialisation in EER_JSON.get_specialisations(eer_model):
        eer_supertype_name =  EER_JSON.get_supertype(specialisation)
        eer_subtype_names = EER_JSON.get_subtypes(specialisation)
        arm_supertype_name = eer_supertype_name
        arm_super_type = ARM_JSON.get_relation_type(arm_model, arm_supertype_name)

        subtype_relation_types = list()
        for eer_subtype_name in eer_subtype_names:
            arm_relation = ARM_JSON.get_relation_type(arm_model, eer_subtype_name)
            subtype_relation_types.append(arm_relation)

            ARM_JSON.add_specialisation_constraint(arm_relation, arm_supertype_name)
            log.append('- added \'isa {:s}\' constraint to relation type \'{:s}\'.'.format(arm_supertype_name, eer_subtype_name))

        if EER_JSON.has_coverage_constraint(specialisation):
            if EER_JSON.get_coverage_constraint(specialisation)==EER_JSON.Coverage.COMPLETE:
                log.append('- added \'covered by ({:s})\' constraint to relation type \'{:s}\'.'.format(ARM_JSON.join(sorted(eer_subtype_names), ', '), arm_supertype_name))
                ARM_JSON.add_coverage_constraint(arm_super_type, eer_subtype_names)

        if EER_JSON.has_disjointness_constraint(specialisation) and EER_JSON.get_disjointness_constraint(specialisation)==EER_JSON.Disjointness.DISJOINT:
            for relation_type in subtype_relation_types:
                relation_type_name = ARM_JSON.relation_type_name(relation_type)
                peer_relation_types = list(eer_subtype_names)
                peer_relation_types.remove(relation_type_name)
                log.append('- added \'disjoint with ({:s})\' to relation type \'{:s}\'.'.format(ARM_JSON.join(sorted(peer_relation_types), ', '), relation_type_name))
                ARM_JSON.add_disjointness_constraint(relation_type, peer_relation_types)
    return


def generate_unique_name(arm_relation_type, participant):
    attribute_names = ARM_JSON.attribute_names(arm_relation_type)
    participant_name = EER_JSON.name(participant).lower()
    if participant_name not in attribute_names:
        # Participant name will do.
        return participant_name
    else:
        # All else has failed, add digits  to participant name
        count = 0
        candidate = participant_name+'_'+str(count)
        while candidate in attribute_names:
            count = count + 1
            candidate = participant_name+'_'+str(count)
        return candidate


def generate_constraint_name(participant, arm_relation_type):
    foreign_key_constraints = ARM_JSON.foreign_key_constraints(ARM_JSON.constraints(arm_relation_type))
    constraint_names = { ARM_JSON.constraint_name(constraint) for constraint in foreign_key_constraints }
    if EER_JSON.has_role(participant) and EER_JSON.get_role(participant) not in constraint_names:
        return EER_JSON.get_role(participant)
    else:
        fk_constraint_number = len(foreign_key_constraints)
        constraint_name = 'fk_const'+str(fk_constraint_number)
        return constraint_name


def map_to_foreign_key(relationship_type, fk_participant, target_participant, arm_model, log):
    log.append("Binary relationship \'{:s}\' between \'{:s}\' and \'{:s}\' mapped to foreign key constraint in ARM relation type \'{:s}\'."\
               .format(EER_JSON.name(relationship_type), EER_JSON.name(fk_participant), EER_JSON.name(target_participant), EER_JSON.name(fk_participant)))
    arm_relation_type = ARM_JSON.get_relation_type(arm_model, EER_JSON.name(fk_participant))
    fk_attribute_name = generate_unique_name(arm_relation_type, target_participant)
    ARM_JSON.add_attribute(arm_relation_type, fk_attribute_name, ARM_JSON.OID)

    # constraint_name = generate_constraint_name(relationship_type, arm_relation_type)
    constraint_name = EER_JSON.name(relationship_type) # Check, may have wrong participant...
    ARM_JSON.add_foreign_key_constraint(arm_relation_type, constraint_name.lower(), fk_attribute_name, EER_JSON.name(target_participant))

    if EER_JSON.has_attributes(relationship_type):
        log.append('- transferred attributes to ARM relation type \'{:s}\'.'.format(EER_JSON.name(fk_participant)))
        transfer_attributes(relationship_type, arm_relation_type, log, EER_JSON.name(relationship_type).lower())

    if EER_JSON.has_participation(fk_participant) and EER_JSON.get_participation(fk_participant)==EER_JSON.Participation.WEAK:
        # This is an identifying relationship for the fk_participant i.e. it is a weak entity
        log.append("- this is an identifying relationship for \'{:s}\'.".format(EER_JSON.name(fk_participant)))
        pathfd_constraints = ARM_JSON.pathfd_constraints(ARM_JSON.constraints(arm_relation_type))
        assert(len(pathfd_constraints)==1) # There shouldn't be more than one identifying pathfd constraint

        pathfd_constraint = pathfd_constraints[0]
        ARM_JSON.add_constraint_path(pathfd_constraint, ARM_JSON.create_simple_path(fk_attribute_name))
        log.append("- set the identifying pathfd constraint for \'{:s}\' to \'{:s}\'.".format(EER_JSON.name(fk_participant), ARM_JSON.constraint_to_string(pathfd_constraint)))

    if EER_JSON.has_min_cardinality(fk_participant):
        log.append('- explicit minimum cardinality \'{:s}\' for \'{:s}\' lost in mapping to ARM.'.format(EER_JSON.get_min_cardinality(fk_participant), EER_JSON.name(fk_participant)))
    if EER_JSON.has_max_cardinality(fk_participant):
        log.append('- explicit maximum cardinality \'{:s}\' for \'{:s}\' lost in mapping to ARM.'.format(EER_JSON.get_max_cardinality(fk_participant), EER_JSON.name(fk_participant)))
    if EER_JSON.has_min_cardinality(target_participant):
        log.append('- minimum cardinality \'{:s}\' for \'{:s}\' lost in mapping to ARM.'.format(EER_JSON.get_min_cardinality(target_participant), EER_JSON.name(target_participant)))
    if EER_JSON.has_max_cardinality(target_participant):
        log.append('- maximum cardinality \'{:s}\' for \'{:s}\' lost in mapping to ARM.'.format(EER_JSON.get_max_cardinality(target_participant), EER_JSON.name(target_participant)))

    return


def transfer_binary_relationship(relationship_type, arm_model, log):
    assert(EER_JSON.is_binary(relationship_type))
    participants = EER_JSON.get_participants(relationship_type)
    # Check if anything has cardinality...
    # Drop foreign key in to relation wth mandatory participation by preference (to avoid foreign key nulls)
    if EER_JSON.cardinality_is_one(participants[0]) and not EER_JSON.optional_participant(participants[0]):
        map_to_foreign_key(relationship_type, participants[0], participants[1], arm_model, log)
    elif EER_JSON.cardinality_is_one(participants[1]) and not EER_JSON.optional_participant(participants[1]):
        map_to_foreign_key(relationship_type, participants[1], participants[0], arm_model, log)
    elif EER_JSON.cardinality_is_one(participants[0]) and FOREIGN_KEY_NULL:
        map_to_foreign_key(relationship_type, participants[0], participants[1], arm_model, log)
    elif EER_JSON.cardinality_is_one(participants[1]) and FOREIGN_KEY_NULL:
        map_to_foreign_key(relationship_type, participants[1], participants[0], arm_model, log)
    else:
        transfer_relationship(relationship_type, arm_model, log)


def transfer_relationship(relationship_type, arm_model, log):
    relationship_name = EER_JSON.name(relationship_type)
    log.append("Relationship type \'{:s}\' mapped to ARM relation type \'{:s}\'.".format(relationship_name, relationship_name))
    arm_relation_type = ARM_JSON.add_relation_type(arm_model, relationship_name)
    # include 'self'
    log.append('- added \'self\' attribute.')
    ARM_JSON.add_attribute(arm_relation_type, ARM_JSON.SELF, ARM_JSON.OID)

    if EER_JSON.has_attributes(relationship_type):
        log.append('- transferred attributes')
        transfer_attributes(relationship_type, arm_relation_type, log)

    foreign_key_names = set()

    for participant in EER_JSON.get_participants(relationship_type):
        assert(not EER_JSON.has_participation(participant) or EER_JSON.get_participation((participant)) != EER_JSON.Participation.WEAK)
        attribute_name = generate_unique_name(arm_relation_type, participant)
        attribute_data_type= ARM_JSON.OID
        ARM_JSON.add_attribute(arm_relation_type, attribute_name, attribute_data_type)

        constraint_name = generate_constraint_name(participant, arm_relation_type)
        ARM_JSON.add_foreign_key_constraint(arm_relation_type, constraint_name.lower(), attribute_name, EER_JSON.name(participant))
        log.append('- added foreign key attribute and constraint for \'{:s}\'.'.format(EER_JSON.name(participant)))
        foreign_key_names.add(attribute_name)

        if EER_JSON.has_min_cardinality(participant):
            log.append('- minimum cardinality \'{:s}\' for \'{:s}\' lost in mapping to ARM.'.format(EER_JSON.get_min_cardinality(participant), EER_JSON.name(participant)))

        if EER_JSON.has_max_cardinality(participant):
            log.append('- maximum cardinality \'{:s}\' for \'{:s}\' lost in mapping to ARM.'.format(EER_JSON.get_max_cardinality(participant), EER_JSON.name(participant)))

    log.append('- added \'primary key (self)\' constraint.')
    ARM_JSON.add_primary_key_constraint(arm_relation_type, ARM_JSON.SELF)
    log.append('- added \'pathfd({:s})->self\' constraint created from foreign key(s).'.format(ARM_JSON.join(sorted(foreign_key_names), ', ')))
    ARM_JSON.add_pathfd_constraint(arm_relation_type, ARM_JSON.create_simple_paths(foreign_key_names), ARM_JSON.create_simple_path(ARM_JSON.SELF))
    return


def transfer_relationships(eer_model, arm_model, log):
    for relationship_type in EER_JSON.relationship_types(eer_model):
        if EER_JSON.is_binary(relationship_type):
            transfer_binary_relationship(relationship_type, arm_model, log)
        else:
            transfer_relationship(relationship_type, arm_model, log)


def fetch_pathfd(relation_type_pathfds, subtype_to_super, relation_name):
    supertype_name = subtype_to_super[relation_name]
    if supertype_name not in relation_type_pathfds:
            fetch_pathfd(relation_type_pathfds, subtype_to_super, supertype_name)
    relation_type_pathfds[relation_name] = relation_type_pathfds[supertype_name]
    return

def calculate_disjointness(arm_model, log):
    # Build an initial dictionary of identifying pathfd.
    relation_type_pathfds = dict()
    # Prepare this structure at the same time. It will store disjoint relation sets
    disjointness_dict = {}

    for relation_type in ARM_JSON.get_relation_types(arm_model):
        relation_type_name = ARM_JSON.relation_type_name(relation_type)
        pathfds = ARM_JSON.pathfd_constraints(ARM_JSON.constraints(relation_type))
        disjointness_dict[relation_type_name] = set()
        for pathfd in pathfds:
            if ARM_JSON.is_identifying(pathfd, 'self'):
                # Assuming at most one identifying pathfd in relation declaration.
                # Subtype relations get theirs from their supertype.
                paths = ARM_JSON.constraint_paths(pathfd)
                # Qualify paths with relation name to avoid name clashes.
                qualified_paths = set()
                for path in paths:
                    qualified_paths.add(relation_type_name+'.'+ARM_JSON.attribute_name(path))
                relation_type_pathfds[relation_type_name]=frozenset(qualified_paths)
            break

    # Now use specialisation relations to propagate paths from super type relations to sub type relations.
    # first build a dictionary mapping sub types to supertypes
    subtype_to_super = dict()
    for relation_type in ARM_JSON.get_relation_types(arm_model):
        relation_type_name = ARM_JSON.relation_type_name(relation_type)
        specialisation_constraints = ARM_JSON.specialisation_constraints(ARM_JSON.constraints(relation_type))
        for specialisation_constraint in specialisation_constraints:
            supertype_name = ARM_JSON.constraint_supertype(specialisation_constraint)
            if relation_type_name not in subtype_to_super:
                subtype_to_super[relation_type_name] = supertype_name

    # Now propagate pathfds
    for relation_type in ARM_JSON.get_relation_types(arm_model):
        relation_type_name = ARM_JSON.relation_type_name(relation_type)
        if relation_type_name not in relation_type_pathfds:
            # The relation does not have a pathfd, must be a subtype.
            assert(relation_type_name in subtype_to_super)
            fetch_pathfd(relation_type_pathfds, subtype_to_super, relation_type_name)


    # And now, calculate disjointness sets.
    relation_type_names = list(relation_type_pathfds.keys())
    for i in range(len(relation_type_names)-1):
        for j in range(i+1, len(relation_type_names)):
            relation_type_A = relation_type_names[i]
            relation_type_B = relation_type_names[j]
            if not relation_type_pathfds[relation_type_A]==relation_type_pathfds[relation_type_B]:
                disjointness_dict[relation_type_A].add(relation_type_B)
                disjointness_dict[relation_type_B].add(relation_type_A)

    # Finally, push disjointness constraints into the relation types.
    for relation_type in ARM_JSON.get_relation_types(arm_model):
        relation_type_name = ARM_JSON.relation_type_name(relation_type)
        if disjointness_dict[relation_type_name]:
            log.append('Added \'disjoint with ({:s})\' constraint to relation type \'{:s}\'.'.format(ARM_JSON.join(sorted(disjointness_dict[relation_type_name]), ', '), relation_type_name))
            ARM_JSON.add_disjointness_constraint(relation_type, disjointness_dict[relation_type_name])

    return

def transform(eer_model):
    log = ["TRANFORMATION REPORT"]
    arm_model = ARM_JSON.create_arm_model()
    ARM_JSON.add_domain(arm_model, ARM_JSON.OID, ARM_JSON.DomainType.ABSTRACT)

    # Transfer entity types
    transfer_entity_types(eer_model, arm_model, log)
    # Transfer specialisations.
    if EER_JSON.has_specialisations(eer_model):
        transfer_specialisations(eer_model, arm_model, log)

    # Transfer relationships
    if EER_JSON.has_relationship_types(eer_model):
        transfer_relationships(eer_model, arm_model, log)

    # Calculate disjointness
    calculate_disjointness(arm_model, log)

    # plug in the log
    ARM_JSON.add_notes(arm_model, log)
    return arm_model
